/*****************************************************************************
    process.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2005 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
#include "auto.h"
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif
#include <sys/wait.h>
#include <signal.h>
#include <pwd.h>
#include <grp.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#include "globals.h"
#include "netio.h"
#include "util.h"
#include "access.h"
#include "logfun.h"
#include "readconfig.h"
#include "signames.h"
#include "ipc.h"
#include "process.h"

#ifndef HAVE_SETEUID
static uid_t base_uid;
static gid_t base_gid;
#endif


int setCore(void)
{
    struct rlimit rl;

    rl.rlim_cur = RLIM_INFINITY;
    rl.rlim_max = RLIM_INFINITY;
    if ((setrlimit(RLIMIT_CORE, &rl))) {
        LOG_MSG1 (LOG_ERR, _("Cannot set core dump limit: %m"));
        return 1;
    }

    return 0;
}

static void dumpProcList(void)
{
#define ITEMSIZE 32
    int count;
    dlp_t *dlp;
    char *dbuf = NULL, *dbp;
    int dbuflen, len=0;

    SIGSTARTBLOCK(SIGCHLD);

    count = getPidCount(procListH);

    if (count) {
        dbuflen = (ITEMSIZE*count)+1;
        if ((dbuf = (char *)Malloc(dbuflen)) == NULL) {
            errNoDynmem;
            return;
        }

        *dbuf = '\0';
        for(dlp=NULL, dbp=dbuf; (dlp = getNextPid(procListH, dlp)) != NULL; ) {
            snprintf(dbp, dbuflen, "%u(%d)%s ", dlp->pid, dlp->mark,
                     dlp->filter ? "F" : "");
            len      = strlen(dbuf);
            dbuflen -= len;
            dbp     += len;
        }
        SIGENDBLOCK(SIGCHLD);

        LOG_MSG4 (LOG_INFO, _p("%d active subprocess: %s", 
                               "%d active subprocesses: %s", count),
                            count, dbuf);
        Free(dbuf);
    }
    else {
        SIGENDBLOCK(SIGCHLD);
        LOG_MSG4 (LOG_INFO, _("No active subprocess"));
    }

}

/*
	everything we need to deal with signals
*/

static char *findSignal (int sigNo)
{
    int i;

    for (i = 0; i < (int)(sizeof(signalNames)/sizeof(signalNames[0])); i++)
        if (signalNames[i].sigNo == sigNo)
            return signalNames[i].sigNam;
    return "???";
}


static void sigHandler (int sigNo)
{
    pid_t child;
    int status, known=1;
#ifdef HAVE_BACKTRACE
#define BTSIZE 64
    static void *array[BTSIZE];
    int aentries;
    static char *sigsevtxt = "Segment violation -- backtrace follows\n";
#endif

    LOG_MSG6 (LOG_DEBUG, "Received Signal %d", sigNo);

    switch (sigNo) {
        case SIGUSR1:
            if (msgLevel < 6)
                msgLevel++; MF_msgLevel = MF_RUN;
                LOG_MSG3 (LOG_NOTICE, _("Message level raised to %d"), msgLevel);
            break;

        case SIGUSR2:
            if (msgLevel > 0)
                msgLevel--; MF_msgLevel = MF_RUN;
                LOG_MSG3 (LOG_NOTICE, _("Message level lowered to %d"), msgLevel);
            break;

        case SIGHUP:
            if (parentPid == getpid()) {
                LOG_MSG3 (LOG_NOTICE, _("Config refresh forced"));
                forceConfig = 1;
            }
            break;

        case SIGINT:
            if (!spMode) {
                if (parentPid == getpid()) {
                    LOG_MSG3 (LOG_NOTICE, _("Config dump to '%s' initiated"),
                              dumpFile);
                    doDump = 1;
                }
                else {
                    doTrace = 1;
                }
                break;
            }			/* fall through to SIGTERM if in spMode */
				/* this enables Ctrl-C termination	*/

        case SIGTERM:
            LOG_MSG3 (LOG_NOTICE, _("Abort request"));
            isAborted = 1;
            longjmp (abortBuf, sigNo);

        case SIGQUIT:		/* used to silently reap a child */
            LOG_MSG4 (LOG_INFO, _("Silent abort request"));
            isAborted = 1;
            if (!(silentDeny & SD_KEEP)) silentDeny = SD_DENY;
            longjmp (abortBuf, sigNo);

        case SIGCONT:		/* signal catched, but ignored	*/
            break;		/* used to wake up the sibling	*/

        case SIGPIPE:
            LOG_MSG3 (LOG_ERR, _("Pipe to filter broken"));
            isAborted = 1;
            longjmp (abortBuf, sigNo);

        case SIGCHLD:
            while ((child = wait3 (&status, WNOHANG, NULL)) > 0) {
                if (WIFEXITED(status)) {
                    filterRtc = WEXITSTATUS(status);
                    LOG_MSG4 (LOG_INFO, _("Sibling %d exited, status %d"), 
                              child, filterRtc);
                }
                else if (WIFSIGNALED(status)) {
                    LOG_MSG4 (LOG_INFO, _("Sibling %d killed, signal %d (%s)"), 
                              child, WTERMSIG(status), 
                              findSignal(WTERMSIG(status)));
                    filterRtc = -1; 
                }
                else {
                    LOG_MSG4 (LOG_INFO, _("Sibling %d died"), child);
                    filterRtc = -1; 
                }
                if (deletePid(procListH, child)) {
                    LOG_MSG1 (LOG_ERR, _("Sibling %d not in process list"), 
                              child);
                }
                DO_MSG5 dumpProcList();

                startTimer(&idleTimer);
                if (parentPid != getpid()) 	/* termination of a filter */
                    if (filterStatus == FILTER_RUNNING)
                        filterStatus = FILTER_DONE;
            }
            break;

#ifdef HAVE_BACKTRACE
        case SIGSEGV:
            write (logFD >= 0 ? logFD : 2, sigsevtxt, strlen(sigsevtxt));
            aentries = backtrace (array, BTSIZE);
            backtrace_symbols_fd (array, aentries, logFD >= 0 ? logFD : 2);
            _exit(sigNo);
#endif

        default:
            known = 0;
    }
    knownSignal += known;
}

void bit2sigMask (ulong bitmask, sigset_t *mask)
{
    int i;
    ulong b;

    sigemptyset(mask);

    for (i = 1, b = 1; i < 32; i++, b <<= 1)
        if (b & bitmask)
             sigaddset(mask, i);
}

int signalSetup(ulong mask, int quiet)
{
    int i, n;
    ulong b;
    struct sigaction act;

    memset (&act, 0, sizeof(act));
    act.sa_handler = sigHandler;
    bit2sigMask (mask, &act.sa_mask);

    for (i = 1, n = 0, b = 1; i < 32; i++, b <<= 1) {
        if (b & mask) {
            if(sigaction(i, &act, NULL)) {
                LOG_MSG1 (LOG_ERR, 
                          _("Cannot set up handler for signal %d (%s): %m"), 
                          i, findSignal(i));
                return 1;
            }
            n++;
        }
    }
    if (!quiet)
        LOG_MSG4 (LOG_INFO, _p("%d signal catched","%d signals catched",n),n);
    return 0;
}
       
int signalReset(ulong mask)
{
    int i, n;
    ulong b;
    struct sigaction act;

    memset (&act, 0, sizeof(act));
    act.sa_handler = SIG_DFL;
    bit2sigMask (mask, &act.sa_mask);

    for (i = 1, n = 0, b = 1; i < 32; i++, b <<= 1) {
        if (b & mask) {
            if(sigaction(i, &act, NULL)) {
                LOG_MSG1 (LOG_ERR, 
                          _("Cannot reset handler for signal  %d (%s): %m"), 
                          i, findSignal(i));
                return 1;
            }
            n++;
        }
    }
    LOG_MSG4 (LOG_INFO, _p("%d signal released","%d signals released",n),n);
    return 0;
}

/*
	clean up at exit of processes
	this function is called via atexit() from all terminating
	(sub)processes.

	if not in single process mode, the parent will try and kill
	all siblings and wait for them to die.
*/


void processRundown(void)
{
    int status, isParent;
    pid_t pid, *pp;
    int pidcount = 0;
    int i;
    
    isParent = ((pid = getpid()) == parentPid);

    if (isAborted) {
        if (!silentDeny && (regCon.pid == (int)pid))
            sendError(regCon.req, PE_UNDEF, "Unexpected server abort");

        if (isParent) {
            LOG_MSG1 (LOG_ERR, _("Server terminating on request"));
        }
        else {
            LOG_MSG2 (LOG_WARNING, _("Sibling %d terminating on request"), pid);
        }
    }

    if (!spMode) {

        SIGBLOCK(SIGCHLD, (pp = getPidVec(procListH)));

        if (pp != NULL) {
            pidcount = (int)(*pp++);

            for (i = pidcount; i; i--, pp++) {
                LOG_MSG5 (LOG_DEBUG, "Waiting for sibling %d to terminate",
                           *pp);
                if (isParent || isAborted)
                    kill (*pp, SIGTERM);
#ifdef HAVE_WAIT4
                wait4(*pp, &status, WNOHANG, NULL);
#elif HAVE_WAITPID
                waitpid(*pp, &status, WNOHANG);
#else
#error	Neither wait4() nor waitpid() supported!
#endif

            }
        }
        if ((isParent || isAborted) && pidcount) {
            LOG_MSG5 (LOG_DEBUG, "%d siblings reaped", pidcount);
        }
        if (isParent) {
            unlink (pidFile);
        }
    }
#ifdef HAVE_MEMDEBUG
    xResume();
#endif
}

/*
	put a stamp into a pid file
*/

static int doPidFile(void)
{
    int  fd, rtc;
    char pidNum[32], *pf = pidFile;

    snprintf(pidNum, sizeof(pidNum), "%d\n", getpid());

    if (*pf == '!')
        pf++;

    LOG_MSG5 (LOG_DEBUG, "PID file %s", pf);

    if ((fd = open (pf, O_WRONLY|O_CREAT|O_TRUNC, 0644)) >= 0) {
        rtc = write(fd, pidNum, strlen(pidNum));
        rtc = (rtc != (int)strlen(pidNum));
    }
    else {
       LOG_MSG1 (LOG_ERR, _("Error creating PID file %s: %m"), pf);
       rtc = 1;
    }
    return rtc;
}


int procEnvSetup(void)
{
    parentPid    = getpid();
    processGroup = getpgrp();

    return (doPidFile() || signalSetup(MIN_SET,0));
}

/*
	take all actions necessary to becom a daemon
*/

int becomeDaemon ()
{
    pid_t pid;
#ifndef HAVE_SETSID
    int tty;
#endif

    if (isDaemon)
        return 0;

    pid = fork();

    if (pid < 0) {
        LOG_MSG1 (LOG_ERR, _("Cannot fork: %m - exiting"));
        exit (1);
    }
    else if (pid > 0) {
        parentPid = 0;		/* needed to prevent processRundown()	*/
        exit (0);		/* from killing himself			*/
    }

    chdir(defaultDir);
    if (coreDump)
        setCore();

#ifdef HAVE_SETSID
    processGroup = setsid();

    if ((int)(processGroup) == -1) {
        LOG_MSG1 (LOG_ERR, _("Cannot change process group: %m"));
        return 1;
    }
#else /* HAVE_SETSID */
#ifdef __osf__
    processGroup = setpgrp(getpid(),getpid());
    if ((int)(processGroup) == -1) 	/* in know, this is a hack, but	*/
         processGroup = setpgrp();	/* i can't still figure out	*/
					/* what makes Tru64 UNIX take 	*/
					/* one or the other :-(		*/
#else
#ifdef SETPGRP_VOID
    processGroup = setpgrp();
#else
    processGroup = setpgrp(getpid(),getpid());
#endif
#endif /* __osf__ */

    if ((int)(processGroup) == -1) {
        LOG_MSG1 (LOG_ERR, _("Cannot change process group: %m"));
        return 1;
    }

    for (tty = 0; tty < 3; tty++)
        if (!isaSocketFd(tty))
            close (tty);

#ifdef USE_TIOCNOTTY
    if ((tty = open("dev/tty", O_RDWR)) < 0) {
        LOG_MSG1 (LOG_ERR, _("Cannot access controlling terminal: %m"));
        return 1;
    }   
#ifdef USE_SETOWN
    if (fcntl (tty, F_SETOWN, getpid())) {
        LOG_MSG1 (LOG_ERR, _("Cannot set tty owner: %m"));
        close (tty);
        return 1;
    }
#endif /* USE_SETOWN */   
    if (ioctl (tty, TIOCNOTTY, NULL)) {
        LOG_MSG1 (LOG_ERR, _("Cannot detach from terminal: %m"));
        close (tty);
        return 1;
    }
    close (tty);
#endif /* USE_TIOCNOTTY */
#endif /* HAVE_SETSID */

    if (procEnvSetup())
        return 1;

    isDaemon = 1;
    LOG_MSG4 (LOG_INFO, _("Detached - process group %d"), (int)processGroup);

    return 0;
}


/*
	create an new child process
*/

pid_t sSpawn (int mark, sai_t *peer)
{
    pid_t pid;

    SIGSTARTBLOCK(SIGCHLD);

    pid = fork();

    if (pid < 0) {
        LOG_MSG1 (LOG_ERR, _("Cannot fork: %m - exiting"));
        exit (1);
    }
    else if (pid > 0) {
        LOG_MSG4 (LOG_INFO, _("Forked sibling PID %d"), (int)pid);

        if (addPid(procListH, mark, pid, 0, peer, NULL))
            LOG_MSG1 (LOG_ERR, _("Cannot update process list"));
        DO_MSG5 dumpProcList();
        SIGENDBLOCK(SIGCHLD);
    }
    else {
        if(signalSetup(CHL_SET,1))  /* switch on special handling    */
            exit (1);
        chdir(defaultDir);
        if (coreDump)
            setCore();
    }

    return pid;
}
 

/*
	switch to runtime UID/GID and set new 'root' if desired
*/

void changeIdent(void)
{
    struct passwd *pw;
    struct group  *gr;

    if (!spMode && changeRoot && (rootDir != NULL)) {
        if (chroot(rootDir)) {
            LOG_MSG1 (LOG_ERR, _("Cannot change to root '%s' : %m - exiting"),
                      rootDir);
            exit (1);
        }
    }

#ifdef HAVE_SETEUID
    setegid(childGid);
    seteuid(childUid);
#else
    base_gid = getgid();
    base_uid = getuid();
    setgid(childGid);
    setuid(childUid);
#endif

    pw = getpwuid(childUid);
    gr = getgrgid(childGid);

    if ((geteuid() != childUid) || (getegid() != childGid)) {
        LOG_MSG1 (LOG_ERR, 
                  _("Cannot change to UID/GID %d/%d (%s/%s): %m - exiting"),
                  (int)childUid, (int)childGid,
                  (pw == NULL) ? "" : pw->pw_name,
                  (gr == NULL) ? "" : gr->gr_name);
        exit (1);
    } 
    else {
        LOG_MSG4 (LOG_INFO, _("Changed to UID/GID %d/%d (%s/%s)"),
                  (int)childUid, (int)childGid,
                  (pw == NULL) ? "" : pw->pw_name,
                  (gr == NULL) ? "" : gr->gr_name);
    } 
}

/*
	reset from runtime to real UID/GID
*/

void resetIdent(void)
{
#ifdef HAVE_SETEUID
    seteuid(getuid());
    setegid(getgid());
    if ((geteuid() != getuid()) || (getegid() != getgid())) {
        LOG_MSG1 (LOG_ERR, _("Cannot reset UID/GID to %d/%d : %m - exiting"),
                  (int)getuid(), (int)getgid());
        exit (1);
    }
#else
    setuid(base_uid);
    setgid(base_gid);
    if ((getuid() != base_uid) || (getgid() != base_gid)) {
        LOG_MSG1 (LOG_ERR, _("Cannot reset UID/GID to %d/%d : %m - exiting"),
                  (int)getuid(), (int)getgid());
        exit (1);
    }
#endif

}


/*
	start a filter as a slave process
	and insert it into the i/o channel chain
*/

int forkFilter (struct tftpR_s *request, int *ds )
{
    int dspair[2], slave_in, slave_out, sts, n, nclose, rtc;
    char *path, *argv0;

    if (pipe(dspair)) {
        LOG_MSG1 (LOG_ERR, _("Cannot create pipe to filter: %m"));
        return -1;
    }

/*
	ds points to the file descriptor currently connected
	to the file on disk. May be reading (RT_RRQ) or writing (RT_WRQ)
*/
    switch (request->reqType) {
        case RT_RRQ:
            slave_in  = *ds;		/* filter takes input from file */
            slave_out = dspair[1];	/* and writes to pipe           */
            *ds       = dspair[0];	/* fileio() now reads from pipe */
            nclose    = dspair[1];	/* to be closed by parent	*/
            break;
        case RT_WRQ:
            slave_in  = dspair[0];	/* filter takes input from pipe */
            slave_out = *ds;		/* and writes to disk           */
            *ds       = dspair[1];	/* fileio() now writes to pipe  */
            nclose    = dspair[0];	/* to be closed by parent	*/
            break;
        default:
            LOG_MSG1 (LOG_ERR, _("%s: Illegal argument '%d'"), 
                      __FUNCTION__, request->reqType);
            return -1;
            break;
    }

    DO_MSG5 {
        char *filterargs;
        filterargs = sprintVarr(request->filter);
        if (filterargs != NULL) {
            doLog(LOG_DEBUG, "Filter exec args: %s", filterargs);
            Free(filterargs);
        }
    }

/*	argv[0] still holds the full path to the filter
	we need acopy of this for the <path> argument of exec(2)
	argv[0] will later be replaced by the short name 
*/

    if ((path = Strdup(request->filter->argv[0])) == NULL) {
        errNoDynmem;
        return -1;
    }
    argv0 = baseName(path, 0);

    filterRtc = 0;
    filterStatus = FILTER_RUNNING;

    if (!(sts = fork())) {

/**/    dup2 (slave_in,  0);		/* map stdin/stdout to filter i/o */
/**/    dup2 (slave_out, 1);
/**/
/**/    request->filter->argv[0] = argv0;/* fill truncated path		*/
/**/
/**/    for (n=2; n<32; n++)		/* close all remaining channels */
/**/        close (n);
/**/
/**/    execv (path, request->filter->argv);	/* start the filter	*/
/**/
/**/	/* we should never get here */
/**/
/**/    LOG_MSG1(LOG_ERR, _("Cannot execv '%s': %m"), path);
/**/    exit (1);
    }

/*
	parent process continues here
*/

    close (nclose);	/* close pipe side given to child */

    if (sts < 0) {
        filterStatus = NO_FILTER;
        LOG_MSG1 (LOG_ERR, _("Cannot fork filter: %m"));
        close (*ds);
        rtc = -1;
    }
    else {
        SIGBLOCK(SIGCHLD, (rtc = addPid(procListH, 1, sts, 1, NULL, NULL)));
        if (rtc)
            LOG_MSG1 (LOG_ERR, _("Cannot add filter '%s' to process list"), 
                      argv0); 

        LOG_MSG4 (LOG_INFO, _("Forked filter '%s' PID %d"), argv0, (int)sts); 

        DO_MSG5 SIGBLOCK(SIGCHLD, (dumpProcList()));
    }
    
    Free(path);

    return rtc;
}
/*
	check the process list for a sibling with a specific peer
	and kill it if found.

	this one is used to reap children, that insist retries on peer ports
	that have been closed already and thus reply with 'port unreachable'
*/

void killChild (sai_t *peer)
{
    dlp_t *pp;

    SIGSTARTBLOCK(SIGCHLD);    
    for (pp = NULL; (pp = getNextPid(procListH, pp)) != NULL;) {
        if ((pp->peer.sin_addr.s_addr == peer->sin_addr.s_addr) &&
            (pp->peer.sin_port == peer->sin_port)) {
            LOG_MSG2 (LOG_WARNING, _("Terminating child %d: Port unreachable"),
                      pp->pid);
            kill (pp->pid, SIGQUIT);
            break;
        }
    }
    SIGENDBLOCK(SIGCHLD);
}
/*
        check existence of an old (stale) PID file and delete it.
*/

int checkPidFile (void)
{
    if (*pidFile == '!') {
        unlink (pidFile+1);
        return 0;
    }
    return !access(pidFile,F_OK);
 }


/*
	scan the process list for active children 
	and send each one a status request
*/

void statsRequest(void)
{
    dlp_t *pp;

    for (pp = NULL; (pp = getNextPid(procListH, pp)) != NULL;) {
        LOG_MSG1 (LOG_DEBUG, "Status request to %d", pp->pid);
        kill (pp->pid, SIGINT);
    }
}


void sigBlock(int signal)
{
    sigset_t newset;

    sigemptyset(&newset);
    sigaddset(&newset, signal);
    if (sigprocmask(SIG_BLOCK, &newset, NULL)) {
        LOG_MSG1 (LOG_ERR, "Cannot block signal %d: %m", signal);
    }
    else {
        LOG_MSG6 (LOG_DEBUG, "Signal %d blocked", signal);
    }
    return;
}


void sigUnblock(int signal)
{
    sigset_t newset;

    sigemptyset(&newset);
    sigaddset(&newset, signal);
    if (sigprocmask(SIG_UNBLOCK, &newset, NULL)) {
        LOG_MSG1 (LOG_ERR, "Cannot unblock signal %d: %m", signal);
    }
    else {
        LOG_MSG6 (LOG_DEBUG, "Signal %d unblocked", signal);
    }
    return;
}
